#pragma once

#include "core/debug/exception.h"
#include "core/debug/safe.h"
#include "core/jobs/jobs.h"
#include "core/improc/map2d.h"

#include <string>
#include <stack>

/* rev_polish_job_base job interface */
class rev_polish_job_base : public job_base
{
public:

	// reverse polish notation only have one return value
	virtual const map2d<float>& get(void) const = 0;
};

/* reverse_polish_invalid_reference_exception exception class */
class reverse_polish_invalid_reference_exception : public exception_base
{
public:
	reverse_polish_invalid_reference_exception(const std::string& id)
	{
		this->m_id = id;
	}

	virtual std::string to_string(void) const override
	{
		return "Invalid reference \"" + this->m_id + "\" !";
	}

private:
	std::string m_id;
};

/* rev_polish_exception exception class */
class rev_polish_exception : public exception_base
{
public:
	rev_polish_exception(const std::string& cmd)
	{
		this->m_cmd = cmd;
	}

	virtual std::string to_string(void) const override
	{
		return "Cannot perform reverse polish operation \"" + this->m_cmd + "\" !";
	}

private:
	std::string m_cmd;
};

/* reverse_polish_empty_stack_exception exception class */
class reverse_polish_empty_stack_exception : public exception_base
{
public:
	virtual std::string to_string(void) const override
	{
		return "Reverse polish stack is empty!";
	}
};

// include all operators
#include "rev_polish_job_ref.h"
#include "rev_polish_job_add.h"
#include "rev_polish_job_subtract.h"
#include "rev_polish_job_neg.h"
#include "rev_polish_job_multconst.h"
#include "rev_polish_job_atan2.h"
#include "rev_polish_job_sqrt.h"
#include "rev_polish_job_square.h"
#include "rev_polish_job_divide.h"

// create reverse polish job
static std::shared_ptr<rev_polish_job_base> create_reverse_polish_job(std::vector<std::shared_ptr<rev_polish_job_base>> refs, const std::string& cmd)
{
	// stack based programming implies a stack
	std::stack<std::string> stack;

	// pop from stack
	auto _pop = [&](void)
	{
		if (stack.empty())
			throw_exception(reverse_polish_empty_stack_exception);

		auto value = stack.top();
		stack.pop();

		return value;
	};

	// push on stack
	auto _push = [&](const std::string& value)
	{
		stack.push(value);
	};

	// reference starts with '&'
	auto _is_ref = [](const std::string& id)
	{
		return str_begins_with(id, "&");
	};

	// create new object
	auto _create = [&](std::shared_ptr<rev_polish_job_base> p)
	{
		auto n = refs.size();

		refs.emplace_back(p);

		char szTmp[64];

		sprintf_s(szTmp, "&%zu", n);

		return std::string(szTmp);
	};

	// get object from stack
	auto _get = [&](const std::string& id)
	{
		if (!str_begins_with(id, "&"))
			throw_exception(reverse_polish_invalid_reference_exception, id);

		auto n = MAKE_SIZE_T(atol(id.c_str() + 1));

		if(n >= refs.size())
			throw_exception(reverse_polish_invalid_reference_exception, id);

		return refs[n];
	};

	// browse all tokens
	auto tokens = tokenize(cmd, ' ');

	for (auto& v : tokens)
	{
		// addition
		if (_strcmpi(v.c_str(), "+") == 0)
		{
			auto op1 = _pop();
			auto op2 = _pop();

			if (_is_ref(op1) && _is_ref(op2))
			{
				auto obj1 = _get(op1);
				auto obj2 = _get(op2);

				auto id = _create(std::make_shared<rev_polish_job_add>(obj1, obj2));

				_push(id);
			}
			else
				throw_exception(rev_polish_exception, op1 + " " + v + " " + op2);
		}
		// subtraction
		else if (_strcmpi(v.c_str(), "-") == 0)
		{
			auto op1 = _pop();
			auto op2 = _pop();

			if (_is_ref(op1) && _is_ref(op2))
			{
				auto obj1 = _get(op1);
				auto obj2 = _get(op2);

				auto id = _create(std::make_shared<rev_polish_job_subtract>(obj1, obj2));

				_push(id);
			}
			else
				throw_exception(rev_polish_exception, op1 + " " + v + " " + op2);
		}
		// divide
		else if (_strcmpi(v.c_str(), "/") == 0)
		{
			auto op1 = _pop();
			auto op2 = _pop();

			if (_is_ref(op1) && _is_ref(op2))
			{
				auto obj1 = _get(op1);
				auto obj2 = _get(op2);

				auto id = _create(std::make_shared<rev_polish_job_divide>(obj1, obj2));

				_push(id);
			}
			else
				throw_exception(rev_polish_exception, op1 + " " + v + " " + op2);
		}
		// multiply
		else if (_strcmpi(v.c_str(), "*") == 0)
		{
			auto op1 = _pop();
			auto op2 = _pop();

			if (_is_ref(op1) && !_is_ref(op2))
			{
				auto obj = _get(op1);
				auto value = (float)atof(op2.c_str());

				auto id = _create(std::make_shared<rev_polish_job_multconst_post>(obj, value));

				_push(id);
			}
			else
				throw_exception(rev_polish_exception, op1 + " " + v + " " + op2);
		}
		// atan2
		else if (_strcmpi(v.c_str(), "atan2") == 0)
		{
			auto op1 = _pop();
			auto op2 = _pop();

			if (_is_ref(op1) && _is_ref(op2))
			{
				auto obj1 = _get(op1);
				auto obj2 = _get(op2);

				auto id = _create(std::make_shared<rev_polish_job_atan2>(obj1, obj2));

				_push(id);
			}
			else
				throw_exception(rev_polish_exception, v + " " + op1 + " " + op2);
		}
		// negation
		else if (_strcmpi(v.c_str(), "neg") == 0)
		{
			auto op = _pop();

			if (_is_ref(op))
			{
				auto obj = _get(op);

				auto id = _create(std::make_shared<rev_polish_job_negate>(obj));

				_push(id);
			}
			else
				throw_exception(rev_polish_exception, v + " " + op);
		}
		// sqrt
		else if (_strcmpi(v.c_str(), "sqrt") == 0)
		{
			auto op = _pop();

			if (_is_ref(op))
			{
				auto obj = _get(op);

				auto id = _create(std::make_shared<rev_polish_job_sqrt>(obj));

				_push(id);
			}
			else
				throw_exception(rev_polish_exception, v + " " + op);
		}
		// square
		else if (_strcmpi(v.c_str(), "^2") == 0)
		{
			auto op = _pop();

			if (_is_ref(op))
			{
				auto obj = _get(op);

				auto id = _create(std::make_shared<rev_polish_job_square>(obj));

				_push(id);
			}
			else
				throw_exception(rev_polish_exception, v + " " + op);
		}
		// duplicate
		else if (_strcmpi(v.c_str(), "dup") == 0)
		{
			auto op = _pop();

			_push(op);
			_push(op);
		}
		// drop
		else if (_strcmpi(v.c_str(), "drop") == 0)
		{
			auto op = _pop();
		}
		// swap
		else if (_strcmpi(v.c_str(), "swap") == 0)
		{
			auto a = _pop();
			auto b = _pop();

			_push(b);
			_push(a);
		}
		// strange operation
		else if (_strcmpi(v.c_str(), "over") == 0)
		{
			auto a = _pop();
			auto b = _pop();

			_push(a);
			_push(b);
			_push(a);
		}
		// rotation
		else if (_strcmpi(v.c_str(), "rot") == 0)
		{
			auto a = _pop();
			auto b = _pop();
			auto c = _pop();

			_push(b);
			_push(c);
			_push(a);
		}
		// add rest stack
		else
			_push(v);
	}

	// return top of stack
	return _get(_pop());
}

// create reverse polish job
static std::shared_ptr<rev_polish_job_base> create_reverse_polish_job(const std::vector<map2d<float>>& objects, const std::string& cmd)
{
	// convert objects to ref type
	std::vector<std::shared_ptr<rev_polish_job_base>> refs(objects.size());

	for (size_t i = 0; i < objects.size(); i++)
		refs[i] = std::make_shared<rev_polish_job_ref>(objects[i]);

	// return main function
	return create_reverse_polish_job(refs, cmd);
}